/*
        Copyright (C) 2017 Graham Jones, www.indriid.com

        This file is part of DENIM.

        DENIM is free software: you can redistribute it and/or modify
        it under the terms of the GNU General Public License as published by
        the Free Software Foundation, either version 3 of the License, or
        (at your option) any later version.

        DENIM is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
        GNU General Public License for more details.

        You should have received a copy of the GNU General Public License
        along with DENIM.  If not, see <http://www.gnu.org/licenses/>.
*/

package denim;

import beast.base.inference.State;
import beast.base.inference.parameter.RealParameter;
import beast.base.core.Citation;
import beast.base.core.Description;
import beast.base.core.Input;
import beast.base.evolution.alignment.TaxonSet;
import beast.base.evolution.tree.Node;
import beast.base.evolution.tree.Tree;
import beast.base.evolution.tree.TreeDistribution;
import beast.base.evolution.tree.TreeInterface;
import denim.util.*;

import java.util.*;


/*
 *  Created by Graham Jones on 27/10/2016.
 */



@Description("Main DENIM class. Calculates coalescence-migration probability.")
@Citation(value=
        "Jones GR (2018) " +
                "   Divergence estimation in the presence of incomplete lineage sorting and migration.\n" +
                "   https://academic.oup.com/sysbio/advance-article-abstract/doi/10.1093/sysbio/syy041/5003138.\n" +
                "   Systematic Biology, syy041, https://doi.org/10.1093/sysbio/syy041\n",
        year = 2018,
        firstAuthorSurname = "Jones",
        DOI="10.1093/sysbio/syy041")


public class CoalescenceMigrationDistribution extends TreeDistribution {
    @SuppressWarnings({"CanBeFinal", "WeakerAccess"})
    public Input<List<GtreeWithEmbedding>> gTreeWEsInput =
            new Input<>("geneTree",
                    "All gene trees, with embedding parameters, and ploidy values.",
                    new ArrayList<GtreeWithEmbedding>());


    @SuppressWarnings({"CanBeFinal", "WeakerAccess"})
    public Input<CoalescenceMigrationParameters> coalescenceMigrationParametersInput =
            new Input<>("coalescenceMigrationParameters",
                    "coalescenceMigrationParameters",
                    Input.Validate.REQUIRED);


    // Needed for Beauti template, but not used here
    @SuppressWarnings("UnusedDeclaration")
    public Input<TaxonSet> taxonSetInput =
            new Input<>("taxonset",
                    "set of taxa mapping lineages to species", Input.Validate.REQUIRED);




    class IntervalInBranch {
        int nLins;  // nof Lineages
        double duration;
        int epoch;

        IntervalInBranch(int nLins, double duration, int epoch) {
            this.nLins = nLins;
            this.duration = duration;
            this.epoch = epoch;
        }
    }





    private Bindings bindings;
    private TreeInterface sTree;
    private int nSMCTreeNodes;
    private int nSMCTreeTips;
    private List<GtreeWithEmbedding> gTreesWE;
    private ArrayList<Tree> gTrees;
    private int nGTrees;

    private double [][] lnGammaRatiosTable;

    private CoalescencesMigrations coalsMigs;

    private STreeBranchesEpochs sTreeBranchesEpochs;
    // stores information that depends only on the species tree

    private int [][] lineageCounts;
    // lineageCounts[b][j] is the number of gene tree lineages from locus j at tipward end of stree branch b.
    private int [][] coalCounts;
    // coalCounts[b][j] = number of coalescences in branch b in gtree j.
    private int [][] migArriveCounts;
    // migArriveCounts[b][j] is the number of lineages arriving in branch b in gtree j
    // via migration.
    private int [][] migLeaveCounts;
    // migLeaveCounts[b][j] is the number of lineages leaving branch b in gtree j
    // via migration.
    private ArrayList<ArrayList<ArrayList<IntervalInBranch>>> branchLocusIntervals;
    // branchLocusIntervals.get(n).get(j) is a list of time branchLocusIntervals and lineages counts in stree branch n for locus j
    private double [][] coalIntensities;
    // coalIntensities[b][j]  =  gamma_bj  =
    //      sum_i c_jbi (n_jbi choose 2)(1/p_j)
    // where c_jbi is i'th time interval and n_jbi are numbers of lineages in the interval,
    // and p_j is ploidy.
    private double [][][] epochBranchMigIntensities;
    // epochBranchMigIntensities[e][b][j] is weighted sum of #lineages in branch b during epoch e for locus j
    private int [][][] migFromToCounts;
    // migFromToCounts[b][d][j] is the number of migrations from branch b to d in gene tree j
    private double [] simpleMigIntensities;
    // simpleMigIntensities[j] is the total migration intensity for gene tree j


    // for store, restore, requiresRecalculation
    private boolean sTreeOrGTreeWEIsDirty[];
    private ArrayList<ArrayList<ArrayList<IntervalInBranch>>> stored_branchLocusIntervals;
    private double [][][] stored_epochBranchMigIntensities;


    // debugging
    private boolean debugFlag = Boolean.valueOf(System.getProperty("denim.debug"));
    private int numberofdebugchecks = 0;
    private final static int maxnumberofdebugchecks = 1000000;





    @Override
    public void initAndValidate() {
        super.initAndValidate(); // ??

        gTreesWE = gTreeWEsInput.get();
        nGTrees = gTreesWE.size();

        gTrees = new ArrayList<>(nGTrees);
        for (int j = 0; j < nGTrees; j++) {
            gTrees.add(gTreesWE.get(j).getTree());
        }

        sTree = treeInput.get();
        bindings = Bindings.initialise(sTree, gTrees);
        nSMCTreeNodes = sTree.getNodeCount();
        nSMCTreeTips = sTree.getLeafNodeCount();

        sTreeBranchesEpochs = new STreeBranchesEpochs(sTree);

        int totalGTipCount = 0;
        for (Tree gtree : gTrees) {
            totalGTipCount += gtree.getLeafNodeCount();
        }
        List<InverseGammaComponent> igms = coalescenceMigrationParametersInput.get().popPriorComponentsInput.get();
        //List<InverseGammaComponent> igms = popPriorComponentsInput.get();
        // number of coalescences and migrations is <= 2*totalGTipCount
        lnGammaRatiosTable = new double[igms.size()][2 * totalGTipCount];
        for (int c = 0; c < igms.size(); c++) {
            InverseGammaComponent igc = igms.get(c);
            for (int q = 0; q < lnGammaRatiosTable[c].length; q++) {
                lnGammaRatiosTable[c][q] = lnRatioGammas(igc.getAlpha(), q);
            }
        }
        coalsMigs = new CoalescencesMigrations(sTree, gTreesWE, bindings);

        lineageCounts = new int[nSMCTreeNodes][nGTrees];
        coalCounts = new int[nSMCTreeNodes][nGTrees];
        migLeaveCounts = new int[nSMCTreeNodes][nGTrees];
        migArriveCounts = new int[nSMCTreeNodes][nGTrees];
        branchLocusIntervals = new ArrayList<>(nSMCTreeNodes);
        stored_branchLocusIntervals = new ArrayList<>(nSMCTreeNodes);
        for (int n = 0; n < nSMCTreeNodes; n++) {
            branchLocusIntervals.add(new ArrayList<>(nGTrees));
            stored_branchLocusIntervals.add(new ArrayList<>(nGTrees));
            for (int j = 0; j < nGTrees; j++) {
                branchLocusIntervals.get(n).add(new ArrayList<>());
                stored_branchLocusIntervals.get(n).add(new ArrayList<>());
            }
        }
        coalIntensities = new double[nSMCTreeNodes][nGTrees];
        epochBranchMigIntensities = new double[nSMCTreeTips][nSMCTreeNodes][nGTrees];
        migFromToCounts = new int[nSMCTreeNodes][nSMCTreeNodes][nGTrees];
        simpleMigIntensities = new double[nGTrees];
        stored_epochBranchMigIntensities = new double[nSMCTreeNodes][nSMCTreeNodes][nGTrees];

        // normalize the inverse gamma component weights to sum to 1
        double totalweight = 0.0;
        for (int i = 0; i < igms.size(); i++) {
            totalweight += igms.get(i).getWeight();
        }
        for (int i = 0; i < igms.size(); i++) {
            igms.get(i).normalizeWeight(totalweight);
        }

        sTreeOrGTreeWEIsDirty = new boolean[nGTrees];
    }


    @Override
    public boolean notCloneable() {
        return true;
    }


    @Override
    public double calculateLogP() {
        if (debugFlag) {
            numberofdebugchecks++;
            if (numberofdebugchecks % 1000 == 0) {
               //System.out.println("DEBUGGING");
            }
        }
        RealParameter popPriorScale = coalescenceMigrationParametersInput.get().popPriorScaleInput.get();
        double sigma = popPriorScale.getValue();

        List<GammaComponent> gams = coalescenceMigrationParametersInput.get().migPriorComponentsInput.get();
        if (gams.size() > 1) {
            System.err.println("Multiple components in the prior for migration rates are not yet implemented.");
            throw new RuntimeException("Fatal DENIM error.");
        }
        double migShape = gams.get(0).getShape();
        double migScale = gams.get(0).getScale();

        double relatednessFactor = coalescenceMigrationParametersInput.get().relatednessFactorInput.get();
        double migrationDecayScale = coalescenceMigrationParametersInput.get().migrationDecayScaleInput.get();
        logP = logCoalescenceMigrationProb(getInverseGammaMixture(), sigma, migShape, migScale, relatednessFactor, migrationDecayScale);
        return logP;
    }


    @Override
    public List<String> getArguments() {
        return null;
    }


    @Override
    public List<String> getConditions() {
        return null;
    }


    @Override
    public void sample(final State state, final Random random) {
    }


    @Override
    protected boolean requiresRecalculation() {
        if (sTree.somethingIsDirty()) {
            for (int j = 0; j < nGTrees; j++) {
                sTreeOrGTreeWEIsDirty[j] = true;
            }
        } else {
            for (int j = 0; j < nGTrees; j++) {
                sTreeOrGTreeWEIsDirty[j] = gTreesWE.get(j).somethingIsDirty();
            }
        }
        return true;
    }


    @Override
    public void store() {
        for (int e = 0; e < nSMCTreeTips; e++) {
            for (int b = 0; b < nSMCTreeNodes; b++) {
                System.arraycopy(epochBranchMigIntensities[e][b], 0, stored_epochBranchMigIntensities[e][b], 0, nGTrees);
            }
        }
        for (int b = 0; b < nSMCTreeNodes; b++) {
            for (int j = 0; j < nGTrees; j++) {
                stored_branchLocusIntervals.get(b).get(j).clear();
                stored_branchLocusIntervals.get(b).get(j).addAll(branchLocusIntervals.get(b).get(j));
            }
        }
        super.store();
    }



    @Override
    public void restore() {
        for (int e = 0; e < nSMCTreeTips; e++) {
            for (int b = 0; b < nSMCTreeNodes; b++) {
                System.arraycopy(stored_epochBranchMigIntensities[e][b], 0, epochBranchMigIntensities[e][b], 0, nGTrees);
            }
        }
        for (int b = 0; b < nSMCTreeNodes; b++) {
            for (int j = 0; j < nGTrees; j++) {
                branchLocusIntervals.get(b).get(j).clear();
                branchLocusIntervals.get(b).get(j).addAll(stored_branchLocusIntervals.get(b).get(j));
            }
        }
        super.restore();
    }



    /* ************************************************************************************************* */
    /* ************************************************************************************************* */
    /* ************************************************************************************************* */



    Tree getStree() {
        return (Tree) sTree;
    }

    // used here and by PopSampleStatistic
    InverseGammaMixture getInverseGammaMixture() {
        List<InverseGammaComponent> igms = coalescenceMigrationParametersInput.get().popPriorComponentsInput.get();
        int n = igms.size();
        InverseGammaMixture igm = new InverseGammaMixture(n);
        for (int c = 0; c < n; c++) {
            igm.setWeight(c, igms.get(c).getWeight());
            igm.setAlpha(c, igms.get(c).getAlpha());
            igm.setBeta(c, igms.get(c).getBeta());
        }
        return igm;
    }



    String getAlignmentID(int j) {
        return gTrees.get(j).getTaxonset().alignmentInput.get().getID();
    }



    // returns the height of the species tree or tallest or gene tree. (with migs, can coalesce before stree root)
    double getMaxTreeHeight() {
        double hgt = sTree.getRoot().getHeight();
        for (Tree gt: gTrees) {
            hgt = Double.max(hgt, gt.getRoot().getHeight());
        }
        return hgt;
    }



    // For logging migCounts
    String getGtreeID(int j) {
        return gTreeWEsInput.get().get(j).getID();
    }


    // For logging migCounts
    double getMigrationCount(int j) {
        coalsMigs.updateMigrationsHeightsForOneGTree(j);
        int nSNodes = sTree.getNodeCount();
        int nMigs = 0;
        for (int n = 0; n < nSNodes; n++) {
            ArrayList<STreeBranchIncident> coalMigs = coalsMigs.getCoalMigsForSNodeNrGTree(n, j);
            for (STreeBranchIncident cm : coalMigs) {
                if (cm.getType() == STreeBranchIncident.Type.MIGRATION_OUT) {
                    nMigs += 1;
                }
            }
        }
        return nMigs;
    }


    // for logging migrations in sTree
    void updateMigrationsHeights() {
        for (int j = 0; j < nGTrees; j++) {
            coalsMigs.updateMigrationsHeightsForOneGTree(j);
        }
    }

    // for logging migrations in sTree
    ArrayList<ArrayList<STreeBranchIncident>> getMigrationsInBranch(Node node) {
        ArrayList<ArrayList<STreeBranchIncident>> nodeMigs = new ArrayList<>();
        for (int j = 0; j < nGTrees; j++) {
            ArrayList<STreeBranchIncident> coalMigs = coalsMigs.getCoalMigsForSNodeNrGTree(node.getNr(), j);
            ArrayList<STreeBranchIncident> nodeLocusMigs = new ArrayList<>();
            for (STreeBranchIncident cm : coalMigs) {
                if (cm.getType() == STreeBranchIncident.Type.MIGRATION_OUT  ||
                        cm.getType() == STreeBranchIncident.Type.MIGRATION_IN) {
                    nodeLocusMigs.add(cm);
                }
            }
            nodeMigs.add(nodeLocusMigs);
        }
        return nodeMigs;
    }





    /*-------------------------------------------------------------------------------------------------*/
    /*------------------------------           private              -----------------------------------*/
    /*-------------------------------------------------------------------------------------------------*/




    private static final Comparator<STreeBranchIncident> HEIGHT_TYPE_ORDER =
            (nlc1, nlc2) -> {
                int comp = Double.compare(nlc1.getHeight(), nlc2.getHeight());
                if (comp != 0) {
                    return comp;
                }
                return STreeBranchIncident.compareTypes(nlc1, nlc2);
            };



    private double logCoalescenceMigrationProb(InverseGammaMixture igm,
                                               double popPriorScale, double migShape, double migScale, double relatednessFactor, double migrationDecayScale) {
        for (int j = 0; j < nGTrees; j++) {
             coalsMigs.updateMigrationsHeightsForOneGTree(j);
        }
        sTreeBranchesEpochs.update(maxGteeHeight());
        updateCoalMigCounts();
        updateLineageCounts();
        updateIntervals();
        updateCoalIntensities();
        double logCP = logCoalescentProb(igm, popPriorScale);
        assert Double.isFinite(logCP);
        double logMP;
        if (relatednessFactor >= 0.0  ||  migrationDecayScale >= 0.0) {
            updateEpochBranchIntensities();
            updateMigFromToCounts();
            logMP = logIOGeneralMigrationProb(migShape, migScale, relatednessFactor, migrationDecayScale);
        } else {
            updateSimpleIntensities();
            logMP = logIOSimpleMigrationProb(migShape, migScale);
        }
        assert Double.isFinite(logMP)  || logMP == Double.NEGATIVE_INFINITY;
        return logCP + logMP;
    }




    /******************* for coalescence and migration calculations ***********************/


    private double maxGteeHeight() {
        double maxGtreeHeight = -1.0;
        for (int j = 0; j < nGTrees; j++) {
            maxGtreeHeight = Math.max(maxGtreeHeight, gTrees.get(j).getRoot().getHeight());
        }
        return maxGtreeHeight;
    }



    private void updateCoalMigCounts() {
        for (int j = 0; j < nGTrees; j++) {

            for (int n = 0; n < nSMCTreeNodes; n++) {
                ArrayList<STreeBranchIncident> colsMigs = coalsMigs.getCoalMigsForSNodeNrGTree(n, j);
                coalCounts[n][j] = 0;
                migLeaveCounts[n][j] = 0;
                migArriveCounts[n][j] = 0;
                for (STreeBranchIncident sbi : colsMigs) {
                    switch (sbi.getType()) {
                        case COALESCENCE:
                            coalCounts[n][j] += 1;
                            break;
                        case MIGRATION_IN:
                            migArriveCounts[n][j] += 1;
                            break;
                        case MIGRATION_OUT:
                            migLeaveCounts[n][j] +=1;
                            break;
                    }
                }
            }
        }
    }



    private void updateLineageCounts() {
        for (int j = 0; j < nGTrees; j++) {
            for (int n = 0; n < sTree.getLeafNodeCount(); n++) {
                lineageCounts[n][j] = bindings.nLineagesForBeastTipNrAndGtree(n, j);
            }
            updateSubtreeNLineages(sTree.getRoot(), j);
        }
    }



    private void updateSubtreeNLineages(Node node, int j) {
        int lftNLin;
        Node lftNode = node.getChild(0);
        if (lftNode.isLeaf()) {
            lftNLin = lineageCounts[lftNode.getNr()][j];
        } else {
            updateSubtreeNLineages(lftNode, j);
            lftNLin = lineageCounts[lftNode.getNr()][j];
        }
        int lftNr = lftNode.getNr();
        lftNLin -= coalCounts[lftNr][j];
        lftNLin += (migArriveCounts[lftNr][j] - migLeaveCounts[lftNr][j]);
        assert lftNLin >= 0;

        int rgtNLin;
        Node rgtNode = node.getChild(1);
        if (rgtNode.isLeaf()) {
            rgtNLin = lineageCounts[rgtNode.getNr()][j];
        } else {
            updateSubtreeNLineages(rgtNode, j);
            rgtNLin = lineageCounts[rgtNode.getNr()][j];
        }
        int rgtNr = rgtNode.getNr();
        rgtNLin -= coalCounts[rgtNr][j];
        rgtNLin += (migArriveCounts[rgtNr][j] - migLeaveCounts[rgtNr][j]);
        assert rgtNLin >= 0;

        lineageCounts[node.getNr()][j] = lftNLin + rgtNLin;
    }



    private void updateIntervals() {
        // collect all incidents per locus j. The boundaries (start/end/epochs), plus the
        // locus-specific coalescences and migrations
        for (int j = 0; j < nGTrees; j++) {
            if (!sTreeOrGTreeWEIsDirty[j]) {
                continue;
            }
            for (int b = 0; b < nSMCTreeNodes; b++) {
                branchLocusIntervals.get(b).get(j).clear();
                ArrayList<STreeBranchIncident> incidents = new ArrayList<>();
                incidents.addAll(sTreeBranchesEpochs.getBoundaries(b));
                incidents.addAll(coalsMigs.getCoalMigsForSNodeNrGTree(b, j));
                incidents.sort(HEIGHT_TYPE_ORDER);
                int e = sTreeBranchesEpochs.getInitialEpoch(b);
                int n_jbi = lineageCounts[b][j];
                for (int i = 1; i < incidents.size(); i++) {
                    double h0 = incidents.get(i-1).getHeight();
                    double h1 = incidents.get(i).getHeight();
                    double c_jbi = h1 - h0;
                    branchLocusIntervals.get(b).get(j).add(new IntervalInBranch(n_jbi, c_jbi, e));
                    if (incidents.get(i).getType() == STreeBranchIncident.Type.EPOCH_BOUNDARY) {
                        e = incidents.get(i).getEpoch();
                    }
                    n_jbi += incidents.get(i).getChange();
                }
            }
        }
    }




    /************************* for coalescences ************************************/

    private void updateCoalIntensities() {
        for (int j = 0; j < nGTrees; j++) {
            for (int n = 0; n < nSMCTreeNodes; n++) {
                coalIntensities[n][j] = 0.0;
            }
        }
        for (int j = 0; j < nGTrees; j++) {
            double coalFactor = gTreesWE.get(j).getCoalFactor();
            for (int n = 0; n < nSMCTreeNodes; n++) {
                ArrayList<IntervalInBranch> x = branchLocusIntervals.get(n).get(j);
                double cI = 0.0;
                for (int i = 0; i < x.size(); i++) {
                    cI += x.get(i).duration * 0.5 * x.get(i).nLins * (x.get(i).nLins - 1);
                }
                cI /= coalFactor;
                coalIntensities[n][j] = cI;
            }
        }
    }




    // Uses the counts and intensities to calculate the coalescent-migration log-probability
    // over all SMC-tree branches.
    private double logCoalescentProb(InverseGammaMixture igm, double popPriorScale) {
        double logP = 0.0;

        double [] logCFs = new double[nGTrees];
        for (int j =0; j < nGTrees; j++) {
            logCFs[j] = Math.log(gTreesWE.get(j).getCoalFactor());
        }
        double [] lambdas = igm.getWeights();
        double [] alphas = igm.getAlphas();
        double [] betas = igm.getBetas();
        double [] logProbCpts = new double[lambdas.length];

        for (int n = 0; n < nSMCTreeNodes; n++) {
            int q_b = 0;
            double gamma_b = 0.0;
            double minusLog_r_b = 0.0;
            for (int j= 0; j < nGTrees; j++) {
                q_b += coalCounts[n][j];
                minusLog_r_b += logCFs[j] * coalCounts[n][j];
                gamma_b += coalIntensities[n][j];
            }
            for (int c = 0; c < lambdas.length; c++) {
                logProbCpts[c] = 0.0;
                double sigmabeta_c = popPriorScale * betas[c];
                logProbCpts[c] += Math.log(lambdas[c]);
                logProbCpts[c] += alphas[c] * Math.log(sigmabeta_c);
                logProbCpts[c] -= (alphas[c] + q_b) * Math.log(sigmabeta_c + gamma_b);
                logProbCpts[c] += lnGammaRatiosTable[c][q_b];
            }
            double logP_b = logSumExp(logProbCpts);
            logP_b -= minusLog_r_b;
            logP += logP_b;
        }
        assert Double.isFinite(logP);
        return logP;
    }



    /***************** for calculations for per branch-pair migration rates ******************/



    private double midPointOfOverlap(int b, int d) {
        Node nb = sTree.getNode(b);
        Node nd = sTree.getNode(d);
        double b0 = nb.getHeight();
        double b1 = nb.getParent().getHeight();
        double d0 = nd.getHeight();
        double d1 = nd.getParent().getHeight();
        double t0 = Math.max(b0, d0);
        double t1 = Math.min(b1, d1);
        return  0.5 * (t0 + t1);
    }



    // TODO store, restore migFromToCounts[][][], and skip j if not dirty
    private void updateMigFromToCounts() {
        for (int j = 0; j < nGTrees; j++) {
            for (int b = 0; b < nSMCTreeNodes; b++) {
                for (int d = 0; d < nSMCTreeNodes; d++) {
                    migFromToCounts[b][d][j] = 0;
                }
            }
            for (int b = 0; b < nSMCTreeNodes; b++) {
                for (int d = 0; d < nSMCTreeNodes; d++) {
                    if (!sTreeBranchesEpochs.branchesCanMigrate(b, d)) {
                        continue;
                    }
                    ArrayList<STreeBranchIncident> coalMigs = coalsMigs.getCoalMigsForSNodeNrGTree(b, j);
                    for (STreeBranchIncident cm : coalMigs) {
                        if (cm.getType() == STreeBranchIncident.Type.MIGRATION_OUT  && cm.getDestNr() == d) {
                            migFromToCounts[b][d][j] += 1;
                        }
                    }
                }
            }
        }
    }



    private void updateEpochBranchIntensities() {
        for (int j = 0; j < nGTrees; j++) {
            if (!sTreeOrGTreeWEIsDirty[j]) {
                continue;
            }
            for (int e = 0; e < epochBranchMigIntensities.length; e++) {
                for (int b = 0; b < nSMCTreeNodes; b++) {
                    epochBranchMigIntensities[e][b][j] = 0.0;
                }
            }
            for (int b = 0; b < nSMCTreeNodes; b++) {
                Node nb = sTree.getNode(b);
                if (!nb.isRoot()) {
                    ArrayList<IntervalInBranch> x = branchLocusIntervals.get(b).get(j);
                    for (int i = 0; i < x.size(); i++) {
                        epochBranchMigIntensities[x.get(i).epoch][b][j] += x.get(i).duration * x.get(i).nLins;
                    }
                }
            }
        }
    }




    private double logIOGeneralMigrationProb(double shape, double scale, double relatednessFactor, double migrationDecayScale) {
        double logMP = 0.0;
        sTreeBranchesEpochs.updateBranchPairDists();
        for (int b = 0; b < nSMCTreeNodes; b++) {
            for (int d = 0; d < nSMCTreeNodes; d++) {
                if (sTreeBranchesEpochs.branchesCanMigrate(b,d)) {
                    double bdMigIntensity = 0.0;
                    int bdMigCount = 0;
                    double gammaScale = scale;
                    if (migrationDecayScale > 0) {
                        double tmrca = sTreeBranchesEpochs.getTmrca(b, d);
                        double overlapMidpoint = midPointOfOverlap(b,d);
                        gammaScale *= Math.exp( -(tmrca - overlapMidpoint) / migrationDecayScale );
                    }
                    if (relatednessFactor >= 0) {
                        int dist = sTreeBranchesEpochs.getTopologicalDistance(b, d);
                        assert(dist >= 2.0);
                        gammaScale *= Math.pow(relatednessFactor, dist-2.0);
                    }
                    for (int j = 0; j < nGTrees; j++) {
                        bdMigCount += migFromToCounts[b][d][j];
                    }
                    for (int e = 0; e < sTreeBranchesEpochs.epochCount(); e++) {
                        if (sTreeBranchesEpochs.overlapsInEpoch(b, d, e)) {
                            for (int j = 0; j < nGTrees; j++) {
                                bdMigIntensity += epochBranchMigIntensities[e][b][j];
                            }
                        }
                    }
                    logMP += lnRatioGammas(shape, bdMigCount);
                    double yy = 0.0;
                    if (bdMigCount > 0) { // avoid 0 * -infinity
                        yy += bdMigCount * Math.log(gammaScale);
                    }
                    yy -= (shape + bdMigCount) * Math.log(1.0 + bdMigIntensity * gammaScale);
                    logMP += yy;
                }
            }
        }
        assert Double.isFinite(logMP)  || logMP == Double.NEGATIVE_INFINITY;
        return logMP;
    }




    /************************** for calculations for single migration rate *************************/


    private void updateSimpleIntensities() {
        for (int j = 0; j < nGTrees; j++) {
            simpleMigIntensities[j] = 0.0;
            for (int b = 0; b < nSMCTreeNodes; b++) {
                Node nb = sTree.getNode(b);
                if (!nb.isRoot()) {
                    double smI = 0.0;
                    ArrayList<IntervalInBranch> x = branchLocusIntervals.get(b).get(j);
                    for (int i = 0; i < x.size(); i++) {
                        smI += x.get(i).duration * x.get(i).nLins * (nSMCTreeTips - x.get(i).epoch - 1);
                    }
                    simpleMigIntensities[j] += smI;
                }
            }
        }
    }



    private double logIOSimpleMigrationProb(double shape, double scale) {
        double logMP = 0.0;
        int migCount = 0;
        double migIntensity = 0.0;
        for (int b = 0; b < nSMCTreeNodes; b++) {
                if (numberofdebugchecks == 5  && b == 7) {
                    System.out.print("DEBUGGING\n");
                }
                if (!sTree.getNode(b).isRoot()) {
                    for (int j = 0; j < nGTrees; j++) {
                        migCount += migLeaveCounts[b][j];
                    }
                }
        }
        for (int j = 0; j < nGTrees; j++) {
            migIntensity += simpleMigIntensities[j];
        }

        logMP += lnRatioGammas(shape, migCount);
        double yy = 0.0;
        if (migCount > 0) { // avoid 0 * -infinity
            yy += migCount * Math.log(scale);
        }
        yy -= (shape + migCount) * Math.log(1.0 + migIntensity * scale);
        logMP += yy;
        assert Double.isFinite(logMP)  || logMP == Double.NEGATIVE_INFINITY;
        return logMP;
    }


    /****************************** maths ***********************************************/



    // returns log(Gamma(a+n)/Gamma(a))
    private double lnRatioGammas(double a, int n) {
        double x = 0.0;
        for (int i = 0; i < n; i++) {
            x += Math.log(a + i);
        }
        return x;
    }


    // Returns log(sum_i e^(x_i)) without over/under flow
    private double logSumExp(double x[]) {
        double maxx = Double.NEGATIVE_INFINITY;
        for (double d : x) {
            if (d > maxx) { maxx = d; }
        }
        double sum = 0.0;
        for (double d : x) {
            sum += Math.exp(d-maxx);
        }
        return maxx + Math.log(sum);
    }


    /****************************** debug output ***********************************************/


    private void debugOutputTreesEmbedding() {
        System.err.flush();
        System.out.println("debugOutputTreesEmbedding()" + " numberofdebugchecks " + numberofdebugchecks);
        System.out.flush();
        System.out.println(Misc.allTreesAsText(sTree, gTrees));
        System.out.flush();
        System.out.println(Misc.embeddingParamsAsText(gTreesWE));
        System.out.flush();
    }



    private void debugOutputGtoSmapping() {
        System.err.flush();
        System.out.println("debugOutputGtoSmapping()" + " numberofdebugchecks " + numberofdebugchecks);
        System.out.println(Misc.gToSmappingAsText(coalsMigs.getGTreeNodeNrToSTreeNodeNr()));
        System.out.flush();
    }


    private void debugOutputCountsIntensities() {
        System.err.flush();
        System.out.println("debugOutputCountsIntensities()" + " numberofdebugchecks " + numberofdebugchecks);
        System.out.println(Misc.CountsIntensitiesAsText(coalCounts, migArriveCounts,
                migLeaveCounts, coalIntensities));
        System.out.flush();
    }



}

