/*
        Copyright (C) 2017 Graham Jones, www.indriid.com

        This file is part of DENIM.

        DENIM is free software: you can redistribute it and/or modify
        it under the terms of the GNU General Public License as published by
        the Free Software Foundation, either version 3 of the License, or
        (at your option) any later version.

        DENIM is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
        GNU General Public License for more details.

        You should have received a copy of the GNU General Public License
        along with DENIM.  If not, see <http://www.gnu.org/licenses/>.
*/

package denim.util;

import beast.base.core.Description;
import beast.base.evolution.tree.Node;
import beast.base.evolution.tree.Tree;
import beast.base.evolution.tree.TreeInterface;

import java.util.List;

/**
 * Created by GRJ on 28/05/2017.
 */

@Description("Various static methods for binary trees")

public class TreeMethods {
    static Node siblingOf(Node xN) {
        assert !xN.isRoot();
        Node ancN = xN.getParent();
        if (ancN.getChild(0) == xN) {
            return ancN.getChild(1);
        } else {
            return ancN.getChild(0);
        }
    }

    static Node ancestorBranchAt(Node x, double height) {
        while (!x.isRoot() && x.getParent().getHeight() < height) {
            x = x.getParent();
        }
        return x;
    }

    static Node mrca(Node x, Node y) {
        while (x != y) {
            if (x.getHeight() < y.getHeight()) {
                x = x.getParent();
            } else {
                y = y.getParent();
            }
        }
        return x;
    }


    // the number of branches between x and y
    static int distanceBetweenNodes(Node x, Node y) {
        int d = 0;
        while (x != y) {
            if (x.getHeight() < y.getHeight()) {
                x = x.getParent();
            } else {
                y = y.getParent();
            }
            d += 1;
        }
        return d;
    }


    static Node [] orderedChildren(Node x, int [] lowestL) {
        assert !x.isLeaf();
        Node lftN = x.getChild(0);
        Node rgtN = x.getChild(1);
        assert lowestL[lftN.getNr()] != lowestL[rgtN.getNr()];
        Node [] nodes = new Node[2];
        if (lowestL[lftN.getNr()] < lowestL[rgtN.getNr()] ) {
            nodes[0] = lftN;
            nodes[1] = rgtN;
        } else {
            nodes[0] = rgtN;
            nodes[1] = lftN;
        }
        return nodes;
    }

    // converts general node number 0...(2n-2) to internal node number 0...(n-2)
    static int findInternalNodeIndex(TreeInterface tree, int Nr) {
        // Node.getNr() is documented:
        /**
         * @return number uniquely identifying the node in the tree.
         *         This is a number between 0 and the total number of nodes in the tree
         *         Leaf nodes are number 0 to #leaf nodes -1
         *         Internal nodes are numbered  #leaf nodes  up to #nodes-1
         *         The root node is always numbered #nodes-1
         */
        assert Nr >= tree.getLeafNodeCount();
        assert Nr < tree.getNodeCount();
        return Nr - tree.getLeafNodeCount();
    }

    // There is a method in BEAST's TreeUtils, but it needs a Tree, not a TreeInterface
    // CoalescenceMigrationDistribution extends TreeDistribution, which uses a TreeInterface.
    public static double treeLength(TreeInterface tree) {
        Node [] nodes = tree.getNodesAsArray();
        double length = 0.0;
        for (int n = 0; n < nodes.length - 1; n++) {
            assert nodes[n].getParent() != null;
            length += nodes[n].getParent().getHeight() - nodes[n].getHeight();
        }
        return length;
    }

    public static int numBranchesAtHeight(TreeInterface tree, double height) {
        List<Node> nodes = tree.getInternalNodes();
        int nb = tree.getLeafNodeCount();
        for (Node node : nodes) {
            if (node.getHeight() < height) {
                nb -= 1;
            }
        }
        return nb;
    }

    // This finds part of the tree length that occurs tipwars of a given height.
    // It is used to find the total length of a gene tree lineages within a species tree,
    // where maxHeight is species tree root height.
    public static double treeLengthUpTo(TreeInterface tree, double maxHeight) {
        Node [] nodes = tree.getNodesAsArray();
        double length = 0.0;
        for (int n = 0; n < nodes.length - 1; n++) {
            assert nodes[n].getParent() != null;
            if (nodes[n].getHeight() < maxHeight) {
                double parentH = Math.min(maxHeight, nodes[n].getParent().getHeight());
                length += parentH - nodes[n].getHeight();
            }
        }
        return length;
    }
}
